/* The MazezaM Engine 
 * Copyright (C) 2002-2008 Malcolm Tyrrell
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package core;

import globals.Globals;

import java.awt.Dimension;
import java.awt.Graphics2D;
import java.awt.Insets;
import java.awt.Rectangle;
import java.awt.event.KeyEvent;
import java.awt.event.MouseEvent;
import java.util.LinkedList;
import java.util.Observable;
import java.util.Observer;

import application.NonlevelGraphics;

import tessellation.Direction;
import tessellation.Position;
import tessellation.TessellationManager;

/**
 * Controls access to a single game accross many levels (e.g. Sokoban levels).
 */
public class GameManager extends Observable implements GameManagerInterface,
		Observer {
	private LevelCollection the_levels;
	private int level_index;

	/** For Undoing actions */
	private GameUndoManager[] undo_manager;

	/**
	 * Stores the argument obtained from observables so that it can be passed to
	 * observers if necessary.
	 */
	Object last_change;

	GameManager(LevelCollection lc) {
		the_levels = lc;
		change_level(0);
		undo_manager = new GameUndoManager[lc.size()];
		for (int i = 0; i < lc.size(); i++)
			undo_manager[i] = new GameUndoManager(the_levels.get(i));
	}

	public String[] get_infoText() {
		String[] info = null;
		String t = the_levels.get(level_index).get_title();
		int offset = 0;
		if (t != "")
			offset = 1;
		info = new String[2 + offset];
		info[0] = Globals.text.getString("level_number") + " "
				+ String.valueOf(level_index + 1) + "/"
				+ String.valueOf(the_levels.size());
		info[1] = Globals.text.getString("open_quote") + t
				+ Globals.text.getString("close_quote");
		info[1 + offset] = "( "
				+ the_levels.get(level_index).get_gameInfo().get_title() + " )";
		return info;
	}

	public LevelCollectionInfo get_levelCollectionInfo() {
		return the_levels;
	}

	/**
	 * This should be the only internal level changing function.
	 */
	private void change_level(int index) {
		if (0 <= index && index < the_levels.size()) {
			the_levels.get(level_index).deleteObserver(this);
			level_index = index;
			// Have to update the DirPosManager's tessellation
			TessellationManager.set_tessellation(the_levels.get(level_index)
					.get_gameInfo().get_tessellation());
			the_levels.get(level_index).addObserver(this);
			setChanged();
		}
	}

	public void action_go(int index) {
		if (index != level_index) {
			change_level(index);
			notifyObserversOfChange();
		}
	}

	public void action_goPrevious() {
		change_level(level_index - 1);
		notifyObserversOfChange();
	}

	public void action_retry() {
		the_levels.get(level_index).deleteObserver(this);
		the_levels.reset(level_index);
		the_levels.get(level_index).addObserver(this);
		undo_manager[level_index].reset_list(the_levels.get(level_index));
		sokoban.Solver.restartSolution();
		setChanged();
		notifyObserversOfChange();
	}

	public void action_goNext() {
		change_level(level_index + 1);
		notifyObserversOfChange();
	}

	public void action_undo() {
		if (undo_manager[level_index].can_undo()) {
			the_levels.get(level_index).deleteObserver(this);
			the_levels.set(level_index, undo_manager[level_index].undo());
			the_levels.get(level_index).addObserver(this);
			setChanged();
		}
		notifyObserversOfChange();
	}

	public void action_mouse(MouseEvent e) {
		if (e.getButton() == MouseEvent.BUTTON1) {
			if (the_levels.get(level_index).is_completed()
					|| the_levels.get(level_index).is_failed()) {
				action_okay();
			} else {
				Position pos = the_levels.get(level_index).get_position(
						e.getPoint());
				// First see if this click is trying to select another player.
				if (!applyAction(new PositionSelectionPlayerAction(0, pos))) {
					// Otherwise, consider it a route action
					LinkedList<Direction> dl = the_levels.get(level_index)
							.findRoute(0, pos);
					if (dl != null) {
						for (Direction d : dl) {
							applyAction(new DirectionPlayerAction(0, d));
						}
					}
				}
			}
		}
		notifyObserversOfChange();
	}

	/**
	 * Attempt to apply the action to the level. Does not notify observers.
	 */
	private boolean applyAction(PlayerAction e) {
		/*
		 * attemptAction can cause the level to change, so I need to store the
		 * level which the undoable action should be added.
		 */
		int i = level_index;
		if (e.attemptAction(the_levels.get(i))) {
			undo_manager[i].add_playerAction(e);
			return true;
		} else
			return false;
	}

	/*
	 * Does not notify observers, since the level will update this object if
	 * necessary.
	 */
	public void action_key(KeyEvent e) {
		int c = e.getKeyCode();
                if (e.getKeyChar()=='n') {
		    try {
		        c = sokoban.Solver.nextMove();
		    } catch(java.util.NoSuchElementException nse) {
		        System.out.println("End of solution.");
			return;
		    }
		}
                if (e.getKeyChar()=='s') {
		    System.out.println(the_levels.get(level_index));
		}
		if (e.getModifiersEx() == 0) {
			if (the_levels.get(level_index).is_completed()
					|| the_levels.get(level_index).is_failed()) {
				action_okay();
			} else {
				applyAction(TessellationManager.getPlayerActionFromKey(c));
			}
		}
		notifyObserversOfChange();
	}

	public void action_okay() {
		if (the_levels.get(level_index).is_completed()) {
			action_goNext();
		} else if (the_levels.get(level_index).is_failed()) {
			action_retry();
		} else {
			applyAction(new AdvanceSelectionPlayerAction(0));
		}
		notifyObserversOfChange();
	}

	public boolean can_goPrevious() {
		return level_index > 0;
	}

	public boolean can_retry() {
		return true;
	}

	public boolean can_goNext() {
		return level_index < the_levels.size() - 1;
	}

	public boolean can_undo() {
		return undo_manager[level_index].can_undo();
	}

	public boolean can_abandonWithoutQuery() {
		/**
		 * This regards completed levels as valuable, which would be appropriate
		 * if I supported the saving of progress.
		 */
		for (GameUndoManager u : undo_manager) {
			if (!u.can_abandonWithoutQuery())
				return false;
		}
		return true;
	}

	public boolean can_provideLevelCollectionInfo() {
		return true;
	}

	public void paintPanel(Graphics2D g2, Insets i, Dimension d) {
		the_levels.get(level_index).paintPanel(g2, i, d);
		if (the_levels.get(level_index).is_completed())
			NonlevelGraphics.levelFinished(g2, i, d);
		else if (the_levels.get(level_index).is_failed())
			NonlevelGraphics.levelFailed(g2, i, d);
	}

	/*
	 * OBSERVER NOTIFICATION
	 */

	/**
	 * This GameManager acts as an observer to levels.
	 */
	public void update(Observable o, Object arg) {
		/*
		 * The arg may be a rectangle which is the area to redraw. However, if
		 * the level has completed or failed, then we ignore the rectangle and
		 * use null instead.
		 */
		if (the_levels.get(level_index).is_completed()
				|| the_levels.get(level_index).is_failed()) {
			last_change = null;
		} else if (last_change == null) {
			last_change = arg;
		} else {
			/**
			 * If we haven't notified the observers since the last update, we
			 * need to account for the union of the rectangles encountered so
			 * far.
			 */
			last_change = ((Rectangle) last_change)
					.createUnion((Rectangle) arg);
		}
		setChanged();
	}

	void notifyObserversOfChange() {
		notifyObservers(last_change);
		last_change = null;
	}
}
