/* The MazezaM Engine 
 * Copyright (C) 2002-2008 Malcolm Tyrrell
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package core;

import java.awt.Dimension;
import java.awt.Graphics2D;
import java.awt.Insets;
import java.awt.Rectangle;
import java.awt.event.KeyEvent;
import java.awt.event.MouseEvent;
import java.util.LinkedList;
import java.util.Observable;
import java.util.Observer;

import application.Dialogs;
import application.NonlevelGraphics;

import file.FileManager;
import file.LevelFileParseException;

/**
 * A sort of delegate class providing access to several GameManagers.
 */
public class GameSwitcher extends Observable implements GameSwitcherInterface,
		Observer {
	private GameManager[] game_managers;
	/**
	 * the current game manager. this is -1 if there are no game_managers yet
	 */
	private int game_manager_index;

	private FileManager file_manager;

	public GameSwitcher() {
		game_manager_index = -1;
		file_manager = new FileManager();
	}

	LevelFileParseException the_error;

	/**
	 * Informs the game panel that the selected game has changed
	 */
	public void action_changeGame(int i) {
		if (0 <= game_manager_index) {
			if (0 <= i && i < game_managers.length) {
				game_managers[game_manager_index].deleteObserver(this);
				game_manager_index = i;
				game_managers[game_manager_index].addObserver(this);
				// ensure the current tessellation is updated
				// TessellationManager.set_tessellation(game_managers[game_manager_index].get_gameInfo().get_tessellation());
				setChanged();
				notifyObservers();
			}
		}
	}

	/*
	 * ACCESSORS
	 */

	public String[] get_infoText() {
		if (0 <= game_manager_index)
			return game_managers[game_manager_index].get_infoText();
		else
			return null;
	}

	public LevelCollectionInfo get_levelCollectionInfo() {
		if (0 <= game_manager_index)
			return game_managers[game_manager_index].get_levelCollectionInfo();
		else
			throw new Error("No game selected yet");
	}

	public String[] get_levelCollectionTitles() {
		if (0 <= game_manager_index) {
			String[] sa = new String[game_managers.length];
			for (int i = 0; i < game_managers.length; i++) {
				sa[i] = game_managers[i].get_levelCollectionInfo().get_title();
			}
			return sa;
		} else {
			return new String[0];
		}
	}

	/*
	 * ACTIONS
	 */
	public void action_go(int index) {
		if (0 <= game_manager_index)
			game_managers[game_manager_index].action_go(index);
	}

	public void action_goPrevious() {
		if (0 <= game_manager_index)
			game_managers[game_manager_index].action_goPrevious();
	}

	public void action_retry() {
		if (0 <= game_manager_index)
			game_managers[game_manager_index].action_retry();
	}

	public void action_goNext() {
		if (0 <= game_manager_index)
			game_managers[game_manager_index].action_goNext();
	}

	public void action_undo() {
		if (0 <= game_manager_index)
			game_managers[game_manager_index].action_undo();
	}

	public void action_mouse(MouseEvent e) {
		if (0 <= game_manager_index)
			game_managers[game_manager_index].action_mouse(e);
		else
			action_okay();
	}

	public void action_key(KeyEvent e) {
		if (0 <= game_manager_index) {
			game_managers[game_manager_index].action_key(e);
		} else {
			action_okay();
		}
	}

	public void action_okay() {
		if (0 <= game_manager_index)
			game_managers[game_manager_index].action_okay();
		else if (game_managers != null && 0 < game_managers.length) {
			game_manager_index = 0;
			setChanged();
		} else {
			if (the_error != null) {
				// Only used in Applet mode!!!
				action_appletLoad();
				setChanged();
				// FIXME - this isn't a good place for notification.
				notifyObservers();
			}
		}
	}

	public void action_openFile() {
		LinkedList<LevelCollection> ll = null;
		try {
			ll = file_manager.openFile();
			if (ll != null) {
				changeLevelManagers(ll);
				notifyObservers(get_levelCollectionTitles());
			}
		} catch (LevelFileParseException e) {
			Dialogs.defaultFileLoadFailed();
			game_manager_index = -1;
		}
	}

	public void action_applicationReload() {
		LinkedList<LevelCollection> ll = null;
		ll = file_manager.reloadLastFile();
		if (ll != null) {
			changeLevelManagers(ll);
			notifyObservers(get_levelCollectionTitles());
		}
	}

	public void action_appletLoad() {
		LinkedList<LevelCollection> ll = null;
		try {
			ll = file_manager.loadAppletLevels();
			changeLevelManagers(ll);
			notifyObservers(get_levelCollectionTitles());
		} catch (LevelFileParseException e) {
			the_error = e;
		}
	}

	public void action_openDefaults() {
		LinkedList<LevelCollection> ll = null;
		try {
			ll = file_manager.loadDefaults();
			changeLevelManagers(ll);
			notifyObservers(get_levelCollectionTitles());
		} catch (LevelFileParseException e) {
			// FIXME
			Dialogs.defaultFileLoadFailed();
			game_manager_index = -1;
		}
	}

	private void changeLevelManagers(LinkedList<LevelCollection> ll) {
		if (ll != null) {
			game_managers = new GameManager[ll.size()];
			game_manager_index = 0;
			for (int i = 0; i < ll.size(); i++)
				game_managers[i] = new GameManager(ll.get(i));
			game_managers[game_manager_index].addObserver(this);
			setChanged();
		}
	}

	/*
	 * QUERIES TO THE GAME PANEL (from observers)
	 */
	public boolean can_goPrevious() {
		if (0 <= game_manager_index)
			return game_managers[game_manager_index].can_goPrevious();
		else
			return false;
	}

	public boolean can_retry() {
		if (0 <= game_manager_index)
			return game_managers[game_manager_index].can_retry();
		else
			return false;
	}

	public boolean can_goNext() {
		if (0 <= game_manager_index)
			return game_managers[game_manager_index].can_goNext();
		else
			return false;
	}

	public boolean can_undo() {
		if (0 <= game_manager_index)
			return game_managers[game_manager_index].can_undo();
		else
			return false;
	}

	public boolean can_abandonWithoutQuery() {
		// check all the game_managers to see if there's a level we shouldn't
		// abandon
		for (GameManager g : game_managers) {
			if (!g.can_abandonWithoutQuery())
				return false;
		}
		return true;
	}

	public boolean can_provideLevelCollectionInfo() {
		if (0 <= game_manager_index)
			return game_managers[game_manager_index]
					.can_provideLevelCollectionInfo();
		else
			return false;
	}

	public void paintPanel(Graphics2D g2, Insets i, Dimension d) {
		if (0 <= game_manager_index)
			game_managers[game_manager_index].paintPanel(g2, i, d);
		else if (the_error != null)
			NonlevelGraphics.error_urlError(the_error, g2, i, d);
		else
			NonlevelGraphics.welcomeScreen(g2, i, d);
	}

	/**
	 * OBSERVABLE
	 */
	public void update() {
		setChanged();
		notifyObservers(true);
	}

	public void update(Observable o, Object arg) {
		setChanged();
		if (arg instanceof Rectangle) {
			notifyObservers(arg);
		} else
			notifyObservers();
	}
}