/* The MazezaM Engine 
 * Copyright (C) 2002-2008 Malcolm Tyrrell
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package level;

import java.util.Iterator;
import java.util.LinkedList;

import tessellation.Position;
import tessellation.TessellationManager;

/**
 * A group of shapes.
 */
// In a multiple inheritance world, this would also extend LinkedList
// and there would be a superclass of this and FloorGroup
public class UnitShapeGroup extends UnitShape {
	public UnitShapeGroup() {
		shapepos_list = new UnitShapePosList();
	}

	public char at(Position pos) {
		Iterator<UnitShapePosPair> spi;
		UnitShapePosPair sp;
		char u = Unit.empty;
		for (spi = shapepos_list.iterator(); spi.hasNext();) {
			sp = spi.next();
			u = Unit.overlap(u, sp.shape.at(pos.minus(sp.pos)));
		}
		return u;
	}

	LinkedList<Position> nonEmptyPositions() {
		return shapepos_list.nonEmptyPositions();
	}

	/**
	 * returns the (shape,position) of the shape which owns the object at the
	 * position or null if no shape does.
	 */
	protected UnitShapePosPair owner(Position pos) {
		Iterator<UnitShapePosPair> spi;
		UnitShapePosPair sp;
		for (spi = shapepos_list.iterator(); spi.hasNext();) {
			sp = spi.next();
			if (!sp.shape.empty(pos.minus(sp.pos)))
				return sp;
		}
		return null;
	}

	/**
	 * returns a list of UnitShapePosPairs which are selected by the provided
	 * selector
	 */
	public UnitShapePosList selected(int s) {
		UnitShapePosList uspl = new UnitShapePosList();
		Iterator<UnitShapePosPair> uspi;
		UnitShapePosPair uspp;
		for (uspi = shapepos_list.iterator(); uspi.hasNext();) {
			uspp = uspi.next();
			if (uspp.shape.selector == s)
				uspl.add(uspp);
		}
		return uspl;
	}

	/**
	 * Move the selection onto the next object in the list (or the first
	 * suitable one if none are currently selected - returns the affected
	 * UnitShapeBounds or null if no change.
	 * 
	 * FIXME - that ain't the way to do it!
	 */
	UnitShapeBounds selectNext(int s) {
		UnitShapeBounds usb;
		Iterator<UnitShapePosPair> uspi;
		UnitShapePosPair uspp;
		// first is the first suitably selectable shape
		UnitShapePosPair first = null;
		// current is the currently selected shape
		UnitShapePosPair current = null;
		// next is next suitable shape after current
		UnitShapePosPair next = null;
		for (uspi = shapepos_list.iterator(); uspi.hasNext();) {
			uspp = uspi.next();
			if (uspp.shape.selectable[s] && (first == null))
				first = uspp;
			if ((current != null) && (next == null) && uspp.shape.selectable[s])
				next = uspp;
			if ((current == null) && (uspp.shape.selector == s))
				current = uspp;
		}
		if (current != null) {
			if (next != null) {
				// select the next shape
				current.shape.selector = UnitShape.unselected;
				next.shape.selector = s;
				usb = new UnitShapeBounds(current.shape.unitShapeBounds(),
						current.pos);
				usb.updateBoundsOffset(next.shape.unitShapeBounds(), next.pos);
				return usb;
			} else if (first == current)
				// no change
				return null;
			else {
				// select the first shape
				current.shape.selector = UnitShape.unselected;
				first.shape.selector = s;
				usb = new UnitShapeBounds(current.shape.unitShapeBounds(),
						current.pos);
				usb
						.updateBoundsOffset(first.shape.unitShapeBounds(),
								first.pos);
				return usb;
			}
		} else if (first != null) {
			first.shape.selector = s;
			usb = new UnitShapeBounds(first.shape.unitShapeBounds(), first.pos);
			return usb;
		} else {
			// FIXME returning rubbish bounds, perhaps should return null
			usb = new UnitShapeBounds(TessellationManager.newPosition());
			return usb;
		}
	}

	/**
	 * returns null if the selection failed or there is no change.
	 */
	UnitShapeBounds attempt_selectAt(int s, Position pos) {
		// find currently selected shape
		UnitShapePosPair current = null;
		for (Iterator<UnitShapePosPair> uspi = shapepos_list.iterator(); uspi
				.hasNext();) {
			UnitShapePosPair uspp = uspi.next();
			if (uspp.shape.selector == s) {
				current = uspp;
				break;
			}
		}
		UnitShapePosPair uspp = owner(pos);
		if (uspp != null && uspp.shape.selectable[s]) {
			if (current == uspp)
				return null;
			UnitShapeBounds bounds = new UnitShapeBounds(pos);
			for (Position p : uspp.shape.nonEmptyPositions())
				bounds.updateBounds(p.plus(uspp.pos));
			if (current != null) {
				current.shape.selector = UnitShape.unselected;
				for (Position p : current.shape.nonEmptyPositions())
					bounds.updateBounds(p.plus(current.pos));
			}
			uspp.shape.selector = s;
			return bounds;
		}
		return null;
	}

	public void add(UnitShapePosPair us) {
		shapepos_list.add(us);
	}

	void paintInto(UnitGraphics g, Position rel, int superSelection) {
		Iterator<UnitShapePosPair> spi;
		UnitShapePosPair sp;
		for (spi = shapepos_list.iterator(); spi.hasNext();) {
			sp = spi.next();
			sp.shape.paintInto(g, sp.pos.plus(rel),
					applySuperSelection(superSelection));
		}
	}

	public Object clone() {
		UnitShapeGroup new_instance = (UnitShapeGroup) super.clone();
		UnitShapePosList new_list = new UnitShapePosList();
		Iterator<UnitShapePosPair> uspli;
		for (uspli = shapepos_list.iterator(); uspli.hasNext();) {
			new_list.add((UnitShapePosPair) uspli.next().clone());
		}
		new_instance.shapepos_list = new_list;
		return new_instance;
	}

	/**
	 * The list of shapes and their positions
	 */
	// FIXME This should be protected! - add a accessor to obtain an iterator
	public UnitShapePosList shapepos_list;
}
