/* 
 * SAT-o-ban - solving planning problems by translation to SAT
 * Copyright (C) 2009 Vladimir Klebanov <vladimir@uni-koblenz.de>
 * based on the MazezaM Engine 
 * Copyright (C) 2002-2008 Malcolm Tyrrell
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package sokoban;

import java.io.BufferedReader;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.util.LinkedList;
import java.util.Iterator;
import java.awt.event.KeyEvent;

import java.io.PrintWriter;

import org.sat4j.BasicLauncher;
import org.sat4j.minisat.SolverFactory;
import org.sat4j.specs.ISolver;


public class Solver {

    public static final boolean DEBUG = false;
    public static final String  LEVEL_INPUT = "levels.txt";
    // sat4j or minisat? minisat is faster, but needs to be compiled
    public static final boolean SAT4JAVA = false; 

    /**
     * The cnf clauses storage before it is written to the file.
     * Needed since the number of clauses appears in the dimacs header.
     */
    private StringBuilder cnfWriter = new StringBuilder();

    /**
     * The counter of clauses, which is used for the dimacs header.
     */
    private int cnfCount = 0;

    private int width;
    
    private int height;
    
    private int depth; // # of last state, i.e., depth = #moves+1
    
    /**
     * The counter of variables, which is used for the dimacs header.
     *
     * The variable sequence is W H#1 C#1 H#2 C#2 ... H#depth C#depth  bottom
     * Where H#i is H_1,1 H_2,1 ... H_width,1 ... H_1,height ... H_width,height
     */
    private int vars;

    /**
     * The index of a variable that is always false. This is guaranteed 
     * by a <code>"-<i>bottom</i> 0"</code> clause. This variable is
     * substituted for all variables that point to a field outside of the
     * level.
     */
    private int bottom;
    
    /**
     * The current level in Sokoban level format
     */
    private String[] level;


    // ######## Complete the methods init() and move() ###############
    // ######## You _can_ leave everything else as is  ###############
    

    public Solver(String[] level, int maxWidth, int depth) {
        this.width = maxWidth;
        this.depth = depth;
        if (DEBUG) System.out.println("width="+width+" depth="+depth);
        height = level.length;
        vars = (width * height * (2*depth+1))+1; // the last one is "bottom"
        bottom = vars;
        comment("Bottom literal");
        clause(-bottom);
        this.level = level;
        comment("Problem clauses");

        init();
        move();
        writeout();
	solve();
	interpret();
    }


    private void solve() {
	try {
            if (SAT4JAVA) {
		BasicLauncher<ISolver> launcher = new BasicLauncher<ISolver>(
                    SolverFactory.instance());
		PrintWriter pw = new PrintWriter("out.ms");
		launcher.setLogWriter(pw);
        	launcher.run(new String[] { "in.ms" });
		launcher.displayResult();
		pw.flush();
		pw.close();
	    } else {
		    Process p = Runtime.getRuntime().exec(
		        "minisat in.ms out.ms");
		    p.waitFor();
	    }
	} catch(IOException ioe) {
	    throw new RuntimeException(ioe);
	} catch(InterruptedException ie) {
	    throw new RuntimeException(ie);
	}
    }
    
    static LinkedList<Integer> solutionKeys;
    static Iterator<Integer> it;
    
    public static int nextMove() {
        if (it==null) it = solutionKeys.iterator();
	return it.next();
    }
    
    public static void restartSolution() {
        it=null;
    }


    private void interpret() {
        solutionKeys = new LinkedList<Integer>();
        String solution = "";
	try {
	    BufferedReader msr = new BufferedReader(new FileReader("out.ms"));
	    String res = msr.readLine();
	    while (res.startsWith("c")) res = msr.readLine(); //skip comments
	    if ("s SATISFIABLE".equals(res)) res="SAT"; // translate sat4j
	    if (DEBUG) System.out.println("SOLVER SAYS:\n"+res);
	    if (res.equals("SAT")) {
		// Split into single variables; remove sat4j's 'v' flag
		String[] model = msr.readLine().replace("v ","").split(" ");
		int i=0; int j=0; int fieldnr=0; int t=0; 
                boolean human=false; // human or container field?
		int oldi=0; int oldj=0;

		for (String l : model) {
                    int val = Integer.parseInt(l);
                    i++;
                    if (DEBUG) {
//                      System.out.print(val+" ");
                        if (val>0) System.out.print("+");
                        else System.out.print(".");
                    }
                    if (i==width) {
                        i=0;
                        j++;
                        if (DEBUG) System.out.println();
                    }
                    if (j==height) {
                        j=0;
                        fieldnr++;
                        t=((fieldnr-1) / 2) +1;
                        human = fieldnr % 2 == 1;
                        if (DEBUG) {
                            System.out.print("*** "+t);
                            if (human) System.out.println(" human");
                            else System.out.println(" containers");
                        }
                    }
                    
		    if (human && (val>0)) {
			if (t>1) {
			    if (i>oldi) {
                                solutionKeys.add(KeyEvent.VK_RIGHT);
                                solution += 'r';
                            } else if (i<oldi) {
                                solutionKeys.add(KeyEvent.VK_LEFT);
                                solution += 'l';
                            } else if (j>oldj) {
                                solutionKeys.add(KeyEvent.VK_DOWN);
                                solution += 'd';
                            } else if (j<oldj) {
                                solutionKeys.add(KeyEvent.VK_UP);
                                solution += 'u';
                            } else throw 
                                new IllegalStateException("Illegal move.");
			}
			
			oldi=i;
			oldj=j;
		    }
		}
	    } else solution = "UNSAT";
	} catch(IOException ioe) {
	    throw new RuntimeException(ioe);
	}
        if (DEBUG) System.out.println("\nSOLUTION:");
	System.out.println(solution);
    }


    private int W(int i, int j) {
	if (i < 1 || i > width || j < 1 || j > height) return bottom;
        return i+(j-1)*width;
    }
    
    private int H(int i, int j, int k) {
	if (i < 1 || i > width || j < 1 || j > height) return bottom;
        return (width*height)*(2*k-1)+i+(j-1)*width;
    }

    private int C(int i, int j, int k) {
	if (i < 1 || i > width || j < 1 || j > height) return bottom;
        return (width*height)*(2*k)+i+(j-1)*width;
    }


    /**
     * given a number of indices, concatenate them and append a 0 to produce a
     * dimacs clause which is added to {@link #cnfWriter} and {@link #cnfCount}
     * incremented.
     * 
     * @param vars
     *            one or more int expressions (clause literals)
     */
    private void clause(int... vars) {
	    for (int i = 0; i < vars.length; i++) {
		    if (vars[i] != 0)
			    cnfWriter.append(vars[i]).append(" ");
	    }
	    cnfWriter.append("0\n");
	    cnfCount++;
    }

    
    private void comment(String s) {
        if (SAT4JAVA) return; // sat4java does not like to read comments
        cnfWriter.append("c "+s+"\n");
    }


    private void writeout() {
        try {
            FileWriter fw = new FileWriter("in.ms");
            fw.write("p cnf " + vars + " " + cnfCount + "\n");
            fw.write(""+cnfWriter);
            fw.close();
            if (DEBUG) System.out.println(
                "Done writing dimacs file, calling solver.");
        } catch(IOException ioe) {
            throw new RuntimeException(ioe);
        }
    }
    
   
    // Complete me
    private void init() {
        comment("*** start init");
        for(int j=1; j<=height; j++) {
            for(int i=1; i<=width; i++) {
                if (i>level[j-1].length()) {
                    //This level line is shorter than level width; fill it up
                } else if (level[j-1].charAt(i-1)=='#') {
                } else if (level[j-1].charAt(i-1)=='@') {
                } else if (level[j-1].charAt(i-1)=='$') {
                } else if (level[j-1].charAt(i-1)==' ') {
                } else if (level[j-1].charAt(i-1)=='.') {
                } else if (level[j-1].charAt(i-1)=='*') {
                }
            }
        }
        comment("*** end init");
    }

    
    // Complete me
    private void move() {
    }

}
