/* The MazezaM Engine 
 * Copyright (C) 2002-2008 Malcolm Tyrrell
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package tessellation;

/**
 * A class which encapsulates the notion of direction, but is also used to
 * describe the wall of a unit (eg northwall) and also the vector along that
 * wall, travelling clockwise.
 */
public abstract class Direction {
	int numberOfDirections;

	public boolean equals(Object other) {
		if (other instanceof Direction)
			return dir == ((Direction) other).dir;
		else
			return false;
	}

	/**
	 * Given an int describing a direction, does it match the current direction
	 */
	public boolean is(int otherDir) {
		if ((otherDir < 0) || (otherDir >= numberOfDirections))
			throw new Error("No such direction as " + otherDir);
		return dir == otherDir;
	}

	/**
	 * Returns the direction clockwise from the current direction
	 */
	public Direction next() {
		return TessellationManager.newDirection((dir + 1) % numberOfDirections);
	}

	public Direction prev() {
		return TessellationManager.newDirection((dir - 1) % numberOfDirections);
	}

	/**
	 * Returns the direction clockwise from the current direction, given the
	 * current position (not all directions are available at all positions)
	 */
	public Direction nextNeighbourAt(Position pos) {
		Direction next = next();
		while (!pos.hasNeighbour(next))
			next = next.next();
		return next;
	}

	/**
	 * Returns the direction clockwise from the current direction, given the
	 * current position (not all edges are available at all positions)
	 */
	public Direction nextEdgeAt(Position pos) {
		Direction next = next();
		while (!pos.hasEdge(next))
			next = next.next();
		return next;
	}

	/**
	 * Returns the direction prior to the current direction (going clockwise),
	 * given the current position (not all edges are available at all positions)
	 */
	public Direction prevEdgeAt(Position pos) {
		Direction next = next();
		while (!equals(next.nextEdgeAt(pos))) {
			next = next.nextEdgeAt(pos);
		}
		return next;
	}

	/**
	 * The opposite direction is the direction you would need to travel to
	 * return to a position having followed the current direction (this is
	 * always meaningful since adjacency is always symmetric).
	 */
	public Direction opposite() {
		return TessellationManager.newDirection((dir + (numberOfDirections / 2))
				% numberOfDirections);
	}

	public String toString() {
		/*
		 * FIXME if ((dir<0)||(dir>=numberOfDirections)) throw new Error("No such
		 * direction as "+dir); return getDirectionName(dir);
		 */
		return (String.valueOf(dir));
	}

	protected int dir;
}